################################################################################
# Created       : Martin Foster
# Modified      : 18-Mar-2003
################################################################################
#
# OpenID - Modules which will handle the complexities associated with the
#          OpenID authentication system.
# Copyright (C) 2000-2007  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 4-3526 Wolfe Cres
#                 Halifax, Nova Scotia
#                 B3L 3S2
#
#################################################################################

package Ethereal::OpenID;					# Package name

use Crypt::OpenID;						# Encryption aspect of OpenID
use strict;							# Strict handling

use Ethereal::Web;						# LWP user agent wrapper


################################################################################
# Sub-Routines
################################################################################

#####################
# Constructor
#
# Allows to class to be exported and thus spawn objects as a result

sub new
{
	#####################
	# Data Members
	my $class    = shift;					# Self reference
	my $database = shift;					# Database details
	my $param    = shift;					# Parameter handler


	#####################
	# Program Area

	# OpenID Crypto Handle
	my $oid      = new Crypt::OpenID();

	# Create internal hash
	my $self     = {
		AGENT    => undef,				# WWW User Agent
		CGI      => $param->{'XCGI'},			# CGI interface
		CRYPT    => $oid,				# Encryption handler
		DATABASE => $database,				# Database handle,
		PARAM    => $param,				# Parameter handler
		ASSOC    => 'HMAC-SHA1',			# Default association type
		ENCRYPT  => 1,					# Encryption use
		SIGN     => 64					# Signature block size
	 };

	# Bless $self into designated class
	bless $self, $class;

	# Return self
	return $self;
}

####################
# Consumer Check
#
# This is the aspect of Ethereal Realms which will handle the complexities
# associated with consumer handling

sub Consumer_Check
{
	####################
	# Data members
	my $self     = shift;					# Reference to self
	my $ident    = shift;					# Identification to check
	my $server   = shift;					# Server to check against

	my $cgi      = $self->{'CGI'};				# CGI interface
	my $database = $self->{'DATABASE'};			# Database handle
	my $oid      = $self->{'CRYPT'};			# OpenID handler
	my $param    = $self->{'PARAM'};			# Parameter list generated
	my $web;						# User agent encapsulation

	my $mac;						# Encrypted key

	my $assoc;						# Association handle
	my $reply;						# Reply
	my $response;						# Server response

	my $inline   = '';					# Inline parameters
	my $sep;						# Seperator
	my $url;						# Url to pass back

	my $elem;						# Singular element
	my $pass;						# What to pass on
	my $sig;						# Generated sig

	my @sig;						# Signature elemtns

	####################
	# Program area

	# Get proper ident, server and associate handle
	($ident, $server, $assoc, $mac) = $self->Discover($ident);

	# Check for existence of server response
	if ((defined($param->{'openid.mode'}))
	  && ($param->{'openid.mode'} =~ /id_res/i))
  	{
		# Acception assertion
		if (defined($param->{'openid.identity'}))
		{
			# What to do next
			# Association failed check against server
			if ((defined($param->{'openid.sig'}))
			 && (defined($param->{'openid.invalidate_handle'})))
		 	{
				# Specific data members
				my %send;
				my %get;

				# Key word
				my $key;


				# Find all appropriate keys
				foreach $key (keys(%{$param}))
				{
					# Transfer over as necessary
					$send{$key} = $param->{$key}
					  if ($key =~ /^openid.*/);
				}

				# Adapt accordingly
				$send{'openid.mode'} = 'check_authentication';


				# Call the server
				$web = new Ethereal::Web();

				# With appropriate arguments
				($reply, $response) = $web->Fetch_Post($server, \%send);

				# What is necessary
				if (defined($reply))
				{
					# Figure out what was received
					$self->Parse($reply, \%get);

					# Remove stale handle
					if ((defined($get{'is_valid'}))
					 && (defined($get{'invalidate_handle'}))
					 && ($get{'is_valid'} eq 'true'))
				 	{
						# Remove from database entry
						$database->Quick("DELETE FROM OpenID
							WHERE OpenIDServer=?
							  AND OpenIDAssociate=?",

						  $server,
						  $get{'invalidate_handle'}
						 );

						# Return successful
						return 2;
					}
				}
			}

			# Signature only
			elsif (defined($param->{'openid.sig'}))
			{
				# Initialize
				$pass = '';

				# Safety
				$param->{'openid.signed'} = 'mode,identity,return_to'
				  unless (defined($param->{'openid.signed'}));


				# Form the line
				# Check for signature
				@sig = split(/,/, $param->{'openid.signed'});

				# Seek out and add in
				foreach $elem (@sig)
				{
					# Append accordingly
					$pass .= $elem . ':' . $param->{'openid.' . $elem} . "\n";
				}

				# Get signature
				$sig = $oid->make_sig($pass, $mac, $self->{'SIGN'});

				# Return value as necessary
				return 2 if ($sig eq  $param->{'openid.sig'});
			}
		}

		# Failure is the default
		else
		{
			# We have what we need?
			if (defined($param->{'openid.user_setup_url'}))
			{
				# Print redirection header
				print $cgi->redirect($param->{'openid.user_setup_url'});

				# Return with sufficient notices
				return 1;
			}
		}
	}

	# Initiate connection
	elsif (!defined($param->{'openid.mode'}))
	{
		# Set base params
		$pass  = 'openid.mode'          . '=' . $cgi->escape('checkid_setup') . '&';
		$pass .= 'openid.identity'      . '=' . $cgi->escape($ident) . '&';
		$pass .= 'openid.assoc_handle'  . '=' . $cgi->escape($assoc) . '&';
		$pass .= 'openid.trust_root'    . '=' . $cgi->escape($database->{'SYS'}{'SetOpenIDRoot'}) . '&';

		# Simple Registration Extension
		$pass .= 'openid.sreg.optional' . '=' . $cgi->escape('fullname,email,nickname') . '&';
		$pass .= 'openid.sreg.policy'   . '=' . $cgi->escape($database->{'SYS'}{'SetOpenIDPolicy'}) . '&';


		# Return the process to us
		# Generate self-referencing link
		$url = $cgi->url(-path_info=>1);

		# Generate inlines
		if (defined($param->{'OPENID'}))
		{
			# Encode appropriatly
			$inline = '?SEC=' . $param->EncodeBase64(
				OPENID  => $param->{'OPENID'},
				MORI    => time + 300
			) . '&B64=True';
		}

		# Append accordingly
		$pass .= 'openid.return_to' . '=' . $cgi->escape($url . $inline);

		# Determine seperator
		$sep = ($server =~ /\?/) ? '&' : '?';

		# Redirection line
		print $cgi->redirect($server . $sep . $pass);

		# Successful return
		return 1;
	}

	# Failed
	return 0;
}


####################
# Discover
#
# This part is required to discover if the OpenID component provided is in
# valid and has all of the necessary information to proceede.  Part of this
# process will also be to associate with the OpenID server and pull the shared
# key.

sub Discover
{
	####################
	# Data members
	my $self     = shift;					# Reference to self
	my $ident    = shift;					# OpenID address

	my $assoc;						# Associate handle
	my $mac;						# Mac Key
	my $reply;						# Reply from their server
	my $response;						# Results in server address
	my $server;						# Server address

	my $state    = 'smart';					# Default state for us

	my $oid      = $self->{'CRYPT'};			# OpenID crypto
	my $database = $self->{'DATABASE'};			# Database handle
	my $web;						# User agent encapsulation

	my %send;						# Arguments to pass
	my %get;						# Arguments pulled

	####################
	# Program area

	#
	# Basic fact checking

	# Return if an IP address
	return undef, undef, undef, undef if ($ident =~ /\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3}/);

	# Correct any missing elements from the HTML
	$ident = 'http://' . $ident unless ($ident =~ /^http[s]?:\/\//i);


	# Pull the page
	# Create instance
	$web = new Ethereal::Web();

	# Try and see
	($reply, $response) = $web->Fetch_Get($ident);

	# Fail if something fails
	return undef, undef, undef, undef, $response unless (defined($reply));


	# Start to find information
	# First necessary component is found
	if ($reply =~ /<link.*rel="openid.server".*href="(.*)?".*>/i)
	{
		# Assign appropriatly
		$server = $1;

		# Is there a delegate?
		$ident  = $1
		  if ($reply =~ /<link.*rel="openid.delegate".*href="(.*)?".*>/i);
	}



	# Associate handle verification
	# Check for existing one
	($assoc, $mac) = $database->DataGet("SELECT
		  OpenIDAssociate AS \"OpenIDAssociate\",
		  OpenIDMac       AS \"OpenIDMac\"
		FROM OpenID
		WHERE OpenIDServer=?
		  AND OpenIDType=?
		  AND OpenIDExpires > LOCALTIMESTAMP
		  AND OpenIDState=?",

	  $server,
	  $self->{'ASSOC'},
	  'smart'
	);


	# Need to associate once again
	unless (defined($assoc))
	{
		# Time to contact the server
		# And associate to get a shared secret
		%send = (
			'openid.mode'       => 'associate',
			'openid.assoc_type' => $self->{'ASSOC'}
		);

		# Is encryption an option
		if ($self->{'ENCRYPT'})
		{
			# Generate a key
			$oid->generate_keys(
				Size     => 64,
				Verbose  => 0 
			);

			# Add the additional components
			$send{'openid.session_type'}       = 'DH-SHA1';
			$send{'openid.dh_modulus'}         = $oid->b64_btwoc($oid->{'p'});
			$send{'openid.dh_gen'}             = $oid->b64_btwoc($oid->{'g'});
			$send{'openid.dh_consumer_public'} = $oid->b64_btwoc($oid->{'pub_key'});
		}

		# Call the server
		# With appropriate arguments
		($reply, $response) = $web->Fetch_Post($server, \%send);

		# Fail if something fails
		return undef, undef, undef, undef, $response unless (defined($reply));

		# Figure out what was received
		$self->Parse($reply, \%get);

		# Extract encoded key
		# Only if it is necessary
		$get{'mac_key'} = $oid->extract_shared(\%get)
		  if (defined($get{'enc_mac_key'}));


		# Write and verify
		#

		eval
		{
			# Interval provided with five less minutes to prevent
			# use of expired keys
			$database->Quick("INSERT INTO OpenID
				(OpenIDServer,
				 OpenIDType,
				 OpenIDAssociate,
				 OpenIDMac,
				 OpenIDExpires,
				 OpenIDState)
				VALUES (?,?,?,?,addInterval('$get{'expires_in'} minutes'),?)",

			  $server,
			  $send{'openid.assoc_type'},
			  $get{'assoc_handle'},
			  $get{'mac_key'},
			  'smart'
			 );
		};

		# Failed
		if ($@)
		{
			# Display a warning
			warn("There was an error in attempting to save OpenID state", $@);

			# Return without success
			return undef, undef, undef, undef;
		}


		# Set accordingly
		$assoc = $get{'assoc_handle'};
		$mac   = $get{'mac_key'};
	}
		
	# Return everything
	return $ident, $server, $assoc, $mac;
}


####################
# Parse
#
# Will parse the output of a returned POST file necessary to get extract the
# information within

sub Parse
{
	####################
	# Data members
	my $self   = shift;					# Reference to self
	my $reply  = shift;					# Reply to parse
	my $args   = shift;					# Arguments hash ref

	my $key;						# Key name
	my $val;						# Key value


	####################
	# Program area

	# Figure out what was received
	while ($reply =~ /^(.+?)\:(.+)$/gm)
	{
		# Precautions
		$key = $1;
		$val = $2;

		# Add to the hash
		$args->{$key} = $val;
	}
}


####################
# Random
#
# Generates a random string from a pool of characters pre-determined during
# the setup.  Nothing complex or necessarily 100% secure.

sub Random
{
	####################
	# Data members
	my $self   = shift;					# Reference to self
	my $length = shift;					# Length to generate

	my $digits;						# Pool to draw from
	my $string;						# Generated string


	####################
	# Program area

	# Safety
	$length = 24
	  unless ((defined($length)) && (length($length) < 256));

	# Initialize pool
	$string = '';
	$digits = 'abcdefghijklmnopqrstuvwzyzABCDEFGHIJKLMNOPQRSTUVWZYZ0123456789';

	# Loop through
	for (1..$length)
	{
		# Append accordingly
		$string .= substr($digits, int(rand(62)), 1);
	}

	# Return string
	return $string;
}

1;								# Return true
